==> icecat-5.0.1-2.tar.xz: icecat-5.0.1-2/layout/tables/SpanningCellSorter.h <==
/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
// vim:cindent:ts=4:et:sw=4:
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla's table layout code.
 *
 * The Initial Developer of the Original Code is the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   L. David Baron <dbaron@dbaron.org> (original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
 * Code to sort cells by their colspan, used by BasicTableLayoutStrategy.
 */

#include "pldhash.h"
#include "nsDebug.h"

class nsIPresShell;

/**
 * The SpanningCellSorter is responsible for accumulating lists of cells
 * with colspans so that those cells can later be enumerated, sorted
 * from lowest number of columns spanned to highest.  It does not use a
 * stable sort (in fact, it currently reverses).
 */
class SpanningCellSorter {
public:
    SpanningCellSorter(nsIPresShell *aPresShell);
    ~SpanningCellSorter();

    struct Item {
        PRInt32 row, col;
        Item *next;
    };

    /**
     * Add a cell to the sorter.  Returns false on out of memory.
     * aColSpan is the number of columns spanned, and aRow/aCol are the
     * position of the cell in the table (for GetCellInfoAt).
     */
    PRBool AddCell(PRInt32 aColSpan, PRInt32 aRow, PRInt32 aCol);

    /**
     * Get the next *list* of cells.  Each list contains all the cells
     * for a colspan value, and the lists are given in order from lowest
     * to highest colspan.  The colspan value is filled in to *aColSpan.
     */
    Item* GetNext(PRInt32 *aColSpan);
private:
    nsIPresShell *mPresShell;

    enum State { ADDING, ENUMERATING_ARRAY, ENUMERATING_HASH, DONE };
    State mState;

    // store small colspans in an array for fast sorting and
    // enumeration, and large colspans in a hash table

    enum { ARRAY_BASE = 2 };
    enum { ARRAY_SIZE = 8 };
    Item *mArray[ARRAY_SIZE];
    PRInt32 SpanToIndex(PRInt32 aSpan) { return aSpan - ARRAY_BASE; }
    PRInt32 IndexToSpan(PRInt32 aIndex) { return aIndex + ARRAY_BASE; }
    PRBool UseArrayForSpan(PRInt32 aSpan) {
        NS_ASSERTION(SpanToIndex(aSpan) >= 0, "cell without colspan");
        return SpanToIndex(aSpan) < ARRAY_SIZE;
    }

    PLDHashTable mHashTable;
    struct HashTableEntry : public PLDHashEntryHdr {
        PRInt32 mColSpan;
        Item *mItems;
    };

    static PLDHashTableOps HashTableOps;

    static PLDHashNumber
        HashTableHashKey(PLDHashTable *table, const void *key);
    static PRBool
        HashTableMatchEntry(PLDHashTable *table, const PLDHashEntryHdr *hdr,
                            const void *key);

    static PLDHashOperator
        FillSortedArray(PLDHashTable *table, PLDHashEntryHdr *hdr,
                        PRUint32 number, void *arg);

    static int SortArray(const void *a, const void *b, void *closure);

    /* state used only during enumeration */
    PRUint32 mEnumerationIndex; // into mArray or mSortedHashTable
    HashTableEntry **mSortedHashTable;

    /*
     * operator new is forbidden since we use the pres shell's stack
     * memory, which much be pushed and popped at points matching a
     * push/pop on the C++ stack.
     */
    void* operator new(size_t sz) CPP_THROW_NEW { return nsnull; }
};

==> icecat-12.0.tar.xz: icecat-12.0/layout/tables/SpanningCellSorter.h <==
/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
// vim:cindent:ts=4:et:sw=4:
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla's table layout code.
 *
 * The Initial Developer of the Original Code is the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   L. David Baron <dbaron@dbaron.org> (original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/*
 * Code to sort cells by their colspan, used by BasicTableLayoutStrategy.
 */

#include "pldhash.h"
#include "nsDebug.h"

class nsIPresShell;

/**
 * The SpanningCellSorter is responsible for accumulating lists of cells
 * with colspans so that those cells can later be enumerated, sorted
 * from lowest number of columns spanned to highest.  It does not use a
 * stable sort (in fact, it currently reverses).
 */
class SpanningCellSorter {
public:
    SpanningCellSorter(nsIPresShell *aPresShell);
    ~SpanningCellSorter();

    struct Item {
        PRInt32 row, col;
        Item *next;
    };

    /**
     * Add a cell to the sorter.  Returns false on out of memory.
     * aColSpan is the number of columns spanned, and aRow/aCol are the
     * position of the cell in the table (for GetCellInfoAt).
     */
    bool AddCell(PRInt32 aColSpan, PRInt32 aRow, PRInt32 aCol);

    /**
     * Get the next *list* of cells.  Each list contains all the cells
     * for a colspan value, and the lists are given in order from lowest
     * to highest colspan.  The colspan value is filled in to *aColSpan.
     */
    Item* GetNext(PRInt32 *aColSpan);
private:
    nsIPresShell *mPresShell;

    enum State { ADDING, ENUMERATING_ARRAY, ENUMERATING_HASH, DONE };
    State mState;

    // store small colspans in an array for fast sorting and
    // enumeration, and large colspans in a hash table

    enum { ARRAY_BASE = 2 };
    enum { ARRAY_SIZE = 8 };
    Item *mArray[ARRAY_SIZE];
    PRInt32 SpanToIndex(PRInt32 aSpan) { return aSpan - ARRAY_BASE; }
    PRInt32 IndexToSpan(PRInt32 aIndex) { return aIndex + ARRAY_BASE; }
    bool UseArrayForSpan(PRInt32 aSpan) {
        NS_ASSERTION(SpanToIndex(aSpan) >= 0, "cell without colspan");
        return SpanToIndex(aSpan) < ARRAY_SIZE;
    }

    PLDHashTable mHashTable;
    struct HashTableEntry : public PLDHashEntryHdr {
        PRInt32 mColSpan;
        Item *mItems;
    };

    static PLDHashTableOps HashTableOps;

    static PLDHashNumber
        HashTableHashKey(PLDHashTable *table, const void *key);
    static bool
        HashTableMatchEntry(PLDHashTable *table, const PLDHashEntryHdr *hdr,
                            const void *key);

    static PLDHashOperator
        FillSortedArray(PLDHashTable *table, PLDHashEntryHdr *hdr,
                        PRUint32 number, void *arg);

    static int SortArray(const void *a, const void *b, void *closure);

    /* state used only during enumeration */
    PRUint32 mEnumerationIndex; // into mArray or mSortedHashTable
    HashTableEntry **mSortedHashTable;

    /*
     * operator new is forbidden since we use the pres shell's stack
     * memory, which much be pushed and popped at points matching a
     * push/pop on the C++ stack.
     */
    void* operator new(size_t sz) CPP_THROW_NEW { return nsnull; }
};
